import { assert } from "o1js";
import { parseRawRegex, generateMinDfaGraph, } from "./regexToDfa.js";
const regex = parseRawRegex("[a-z]+");
const graphJson = JSON.parse(generateMinDfaGraph(regex));
const N = graphJson.length;
// Outgoing nodes
const graph = Array.from({ length: N }, () => ({}));
// Incoming Nodes
const revGraph = Array.from({ length: N }, () => []);
// let init_graph: number[][] = Array.from({ length: N }, () => []);
// let init_going_state: number | undefined;
const acceptNodes = new Set();
// Started Rust Adaptation //
// for (let i = 0; i < N; i++) {
//     const node = graphJson[i];
//     for (const [k, v] of Object.entries(node.transition)) {
//         const chars: number[] = Object.values(v);
//         revGraph[k] ??= [];
//         revGraph[k][i] = chars.slice(); // Clone the array
//         if (i === 0) {
//             const index = chars.findIndex(x => x === 94);
//             if (index !== -1) {
//                 init_going_state = i;
//                 revGraph[k][i][index] = 255;
//             }
//             for (const j of (revGraph[k][i] ?? [])) {
//                 if (j === 255) {
//                     continue;
//                 }
//                 init_graph[i] ??= [];
//                 init_graph[i].push(j);
//             }
//         }
//     }
//     if (node.type === "accept") {
//         acceptNodes.add(i);
//     }
// }
// Finished Rust Adaptation //
for (let i = 0; i < N; i++) {
    // graph.push({});
    // revGraph.push([]);
    const currentNode = graphJson[i];
    for (const k in currentNode.transition) {
        const v = currentNode.transition[k];
        graph[i][k] = v;
        revGraph[v].push([k, i]);
    }
    if (currentNode.type === 'accept') {
        acceptNodes.add(i);
    }
}
const acceptNodesArray = Array.from(acceptNodes);
if (acceptNodesArray.length !== 1) {
    throw new Error("Accept nodes length must be exactly 1.");
}
let eq_i = 0;
let lt_i = 0;
let and_i = 0;
let multi_or_i = 0;
let range_checks = [];
let eq_checks = [];
let multi_or_checks1 = [];
let multi_or_checks2 = [];
let lines = [];
lines.push("for (var i = 0; i < num_bytes; i++) {");
lines.push(`\t\tstate_changed[i] = MultiOR(${N - 1});`);
if (acceptNodesArray.includes(0)) {
    throw new Error('0 should not be in accept nodes');
}
for (let i = 1; i < N; i++) {
    const outputs = [];
    for (const [k, prev_i] of revGraph[i]) {
        let vals = new Set(k.split(','));
        const eq_outputs = [];
        const uppercase = new Set("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
        const lowercase = new Set("abcdefghijklmnopqrstuvwxyz");
        const digits = new Set("0123456789");
        if (Array.from(uppercase).every(val => vals.has(val))) {
            // vals.forEach(val => vals.delete(val));
            vals = new Set([...vals].filter(char => !uppercase.has(char)));
            lines.push(`\tlt[${lt_i}][i] = LessThan(8);`);
            lines.push(`\tlt[${lt_i}][i].in[0] <== 64;`);
            lines.push(`\tlt[${lt_i}][i].in[1] <== in[i];`);
            lines.push(`\tlt[${lt_i + 1}][i] = LessThan(8);`);
            lines.push(`\tlt[${lt_i + 1}][i].in[0] <== in[i];`);
            lines.push(`\tlt[${lt_i + 1}][i].in[1] <== 91;`);
            lines.push(`\tand[${and_i}][i] = AND();`);
            lines.push(`\tand[${and_i}][i].a <== lt[${lt_i}][i].out;`);
            lines.push(`\tand[${and_i}][i].b <== lt[${lt_i + 1}][i].out;`);
            eq_outputs.push(['and', and_i]);
            lt_i += 2;
            and_i += 1;
        }
        if (Array.from(lowercase).every(val => vals.has(val))) {
            // vals.forEach(val => vals.delete(val));
            vals = new Set([...vals].filter(char => !lowercase.has(char)));
            lines.push(`\tlt[${lt_i}][i] = LessThan(8);`);
            lines.push(`\tlt[${lt_i}][i].in[0] <== 96;`);
            lines.push(`\tlt[${lt_i}][i].in[1] <== in[i];`);
            lines.push(`\tlt[${lt_i + 1}][i] = LessThan(8);`);
            lines.push(`\tlt[${lt_i + 1}][i].in[0] <== in[i];`);
            lines.push(`\tlt[${lt_i + 1}][i].in[1] <== 123;`);
            lines.push(`\tand[${and_i}][i] = AND();`);
            lines.push(`\tand[${and_i}][i].a <== lt[${lt_i}][i].out;`);
            lines.push(`\tand[${and_i}][i].b <== lt[${lt_i + 1}][i].out;`);
            eq_outputs.push(['and', and_i]);
            lt_i += 2;
            and_i += 1;
        }
        if (Array.from(digits).every(val => vals.has(val))) {
            // vals.forEach(val => vals.delete(val));
            vals = new Set([...vals].filter(char => !digits.has(char)));
            lines.push(`\tlt[${lt_i}][i] = LessThan(8);`);
            lines.push(`\tlt[${lt_i}][i].in[0] <== 47;`);
            lines.push(`\tlt[${lt_i}][i].in[1] <== in[i];`);
            lines.push(`\tlt[${lt_i + 1}][i] = LessThan(8);`);
            lines.push(`\tlt[${lt_i + 1}][i].in[0] <== in[i];`);
            lines.push(`\tlt[${lt_i + 1}][i].in[1] <== 58;`);
            lines.push(`\tand[${and_i}][i] = AND();`);
            lines.push(`\tand[${and_i}][i].a <== lt[${lt_i}][i].out;`);
            lines.push(`\tand[${and_i}][i].b <== lt[${lt_i + 1}][i].out;`);
            eq_outputs.push(['and', and_i]);
            lt_i += 2;
            and_i += 1;
        }
        for (const c of vals) {
            assert(c.length === 1 || c.length === 0);
            lines.push(`\teq[${eq_i}][i] = IsEqual();`);
            lines.push(`\teq[${eq_i}][i].in[0] <== in[i];`);
            lines.push(`\teq[${eq_i}][i].in[1] <== ${c.charCodeAt(0)};`);
            eq_outputs.push(['eq', eq_i]);
            eq_i += 1;
        }
        lines.push(`\tand[${and_i}][i] = AND();`);
        lines.push(`\tand[${and_i}][i].a <== states[i][${prev_i}];`);
        if (eq_outputs.length === 1) {
            lines.push(`\tand[${and_i}][i].b <== ${eq_outputs[0][0]}[${eq_outputs[0][1]}][i].out;`);
        }
        else if (eq_outputs.length > 1) {
            lines.push(`\tmulti_or[${multi_or_i}][i] = MultiOR(${eq_outputs.length});`);
            for (let output_i = 0; output_i < eq_outputs.length; output_i++) {
                lines.push(`\tmulti_or[${multi_or_i}][i].in[${output_i}] <== ${eq_outputs[output_i][0]}[${eq_outputs[output_i][1]}][i].out;`);
            }
            lines.push(`\tand[${and_i}][i].b <== multi_or[${multi_or_i}][i].out;`);
            multi_or_i += 1;
        }
        outputs.push(and_i);
        and_i += 1;
    }
    if (outputs.length === 1) {
        lines.push(`\tstates[i+1][${i}] <== and[${outputs[0]}][i].out;`);
    }
    else if (outputs.length > 1) {
        lines.push(`\tmulti_or[${multi_or_i}][i] = MultiOR(${outputs.length});`);
        for (let output_i = 0; output_i < outputs.length; output_i++) {
            lines.push(`\tmulti_or[${multi_or_i}][i].in[${output_i}] <== and[${outputs[output_i]}][i].out;`);
        }
        lines.push(`\tstates[i+1][${i}] <== multi_or[${multi_or_i}][i].out;`);
        multi_or_i += 1;
    }
}
lines.push("}");
const declarations = [];
if (eq_i > 0) {
    declarations.push(`component eq[${eq_i}][num_bytes];`);
}
if (lt_i > 0) {
    declarations.push(`component lt[${lt_i}][num_bytes];`);
}
if (and_i > 0) {
    declarations.push(`component and[${and_i}][num_bytes];`);
}
if (multi_or_i > 0) {
    declarations.push(`component multi_or[${multi_or_i}][num_bytes];`);
}
declarations.push(`signal states[num_bytes+1][${N}];`);
declarations.push("");
const init_code = [];
init_code.push("for (var i = 0; i < num_bytes; i++) {");
init_code.push("\tstates[i][0] <== 1;");
init_code.push("}");
init_code.push(`for (var i = 1; i < ${N}; i++) {`);
init_code.push("\tstates[0][i] <== 0;");
init_code.push("}");
init_code.push("");
lines = [...declarations, ...init_code, ...lines];
const accept_node = acceptNodesArray[0];
const accept_lines = [""];
accept_lines.push("signal final_state_sum[num_bytes+1];");
accept_lines.push(`final_state_sum[0] <== states[0][${accept_node}];`);
accept_lines.push("for (var i = 1; i <= num_bytes; i++) {");
accept_lines.push(`\tfinal_state_sum[i] <== final_state_sum[i-1] + states[i][${accept_node}];`);
accept_lines.push("}");
accept_lines.push("out <== final_state_sum[num_bytes];");
lines.push(...accept_lines);
// console.log(lines.join("\n"));
/*
pragma circom 2.1.5;

include "@zk-email/zk-regex-circom/circuits/regex_helpers.circom";

// regex: [a-z]+
template Test(msg_bytes) {
    signal input msg[msg_bytes];
    signal output out;

    var num_bytes = msg_bytes+1;
    signal in[num_bytes];
    in[0]<==255;
    for (var i = 0; i < msg_bytes; i++) {
        in[i+1] <== msg[i];
    }

    component lt[2][num_bytes];
    component and[3][num_bytes];
    component multi_or[1][num_bytes];
    signal states[num_bytes+1][2];
    component state_changed[num_bytes];

    states[0][0] <== 1;
    for (var i = 1; i < 2; i++) {
        states[0][i] <== 0;
    }

    for (var i = 0; i < num_bytes; i++) {
        state_changed[i] = MultiOR(1);
        lt[0][i] = LessEqThan(8);
        lt[0][i].in[0] <== 97;
        lt[0][i].in[1] <== in[i];
        lt[1][i] = LessEqThan(8);
        lt[1][i].in[0] <== in[i];
        lt[1][i].in[1] <== 122;
        and[0][i] = AND();
        and[0][i].a <== lt[0][i].out;
        and[0][i].b <== lt[1][i].out;
        and[1][i] = AND();
        and[1][i].a <== states[i][0];
        and[1][i].b <== and[0][i].out;
        and[2][i] = AND();
        and[2][i].a <== states[i][1];
        and[2][i].b <== and[0][i].out;
        multi_or[0][i] = MultiOR(2);
        multi_or[0][i].in[0] <== and[1][i].out;
        multi_or[0][i].in[1] <== and[2][i].out;
        states[i+1][1] <== multi_or[0][i].out;
        state_changed[i].in[0] <== states[i+1][1];
        states[i+1][0] <== 1 - state_changed[i].out;
    }

    component final_state_result = MultiOR(num_bytes+1);
    for (var i = 0; i <= num_bytes; i++) {
        final_state_result.in[i] <== states[i][1];
    }
    out <== final_state_result.out;

    signal is_consecutive[msg_bytes+1][2];
    is_consecutive[msg_bytes][1] <== 1;
    for (var i = 0; i < msg_bytes; i++) {
        is_consecutive[msg_bytes-1-i][0] <== states[num_bytes-i][1] * (1 - is_consecutive[msg_bytes-i][1]) + is_consecutive[msg_bytes-i][1];
        is_consecutive[msg_bytes-1-i][1] <== state_changed[msg_bytes-i].out * is_consecutive[msg_bytes-1-i][0];
    }

    signal is_substr0[msg_bytes][1];
    signal is_reveal0[msg_bytes];
    signal output reveal0[msg_bytes];
    for (var i = 0; i < msg_bytes; i++) {
        is_substr0[i][0] <== 0;
        is_reveal0[i] <== is_substr0[i][0] * is_consecutive[i][1];
        reveal0[i] <== in[i+1] * is_reveal0[i];
    }
}


*/ 
//# sourceMappingURL=genRust.js.map