/* eslint-disable no-redeclare */
/* eslint-disable no-prototype-builtins */
/*global require, exports*/
// import { STRING_PRESELECTOR } from "../src/helpers/constants.ts";
import { minDfa, nfaToDfa, regexToNfa } from "./rawlex.js";
/** This section defines helper regex components -- to edit the regex used, edit the return
 * of the test_regex function.
 * All of the relevant regexes are in the main repo README.
 */
// Helper components
const a2z = "a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z";
const A2Z = "A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z";
const r0to9 = "0|1|2|3|4|5|6|7|8|9";
const alphanum = `${a2z}|${A2Z}|${r0to9}`;
const key_chars = `(${a2z})`;
const catch_all = "(0|1|2|3|4|5|6|7|8|9|a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|!|\"|#|$|%|&|'|\\(|\\)|\\*|\\+|,|-|.|/|:|;|<|=|>|\\?|@|\\[|\\\\|\\]|^|_|`|{|\\||}|~| |\t|\n|\r|\x0b|\x0c)";
const catch_all_without_semicolon = "(0|1|2|3|4|5|6|7|8|9|a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|!|\"|#|$|%|&|'|\\(|\\)|\\*|\\+|,|-|.|/|:|<|=|>|\\?|@|\\[|\\\\|\\]|^|_|`|{|\\||}|~| |\t|\n|\r|\x0b|\x0c)";
const email_chars = `${alphanum}|_|.|-`;
const base_64 = `(${alphanum}|\\+|/|=)`;
const word_char = `(${alphanum}|_)`;
const a2z_nosep = "abcdefghijklmnopqrstuvwxyz";
const A2Z_nosep = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
const a2f_nosep = "abcdef";
const A2F_nosep = "ABCDEF";
const r0to9_nosep = "0123456789";
// TODO: Note that this is replicated code in lexical.js as well
// Note that ^ has to be manually replaced with \x80 in the regex
const escapeMap = { n: "\n", r: "\r", t: "\t", v: "\v", f: "\f" };
let whitespace = Object.values(escapeMap);
const slash_s = whitespace.join("|");
// The test_regex function whose return needs to be edited
// Note that in order to specify some strings in regex, we must use \\ to escape \'s.
// For instance, matching the literal + is represented as \\+.
// However, matching the literal \r (ascii 60) character is still \r
// Matching \ then an r as two characters would be \\r in the js string literal
function test_regex() {
    // let to_from_regex_old = '(\r\n|\x80)(to|from):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>?\r\n';
    // let regex = `\r\ndkim-signature:(${key_chars}=${catch_all_without_semicolon}+; )+bh=${base_64}+; `;
    // let order_invariant_regex_raw = `((\\n|\x80|^)(((from):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>)?|(subject:[a-zA-Z 0-9]+)?|((to):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>)?|(dkim-signature:((a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z)=(0|1|2|3|4|5|6|7|8|9|a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|!|"|#|$|%|&|\'|\\(|\\)|\\*|\\+|,|-|.|/|:|<|=|>|\\?|@|[|\\\\|]|^|_|`|{|\\||}|~| |\t|\n|\r|\x0B|\f)+; ))?)(\\r))+` // Uses a-z syntax instead of | for each char
    let email_address_regex = `([a-zA-Z0-9._%\\+-=]+@[a-zA-Z0-9.-]+)`;
    // ------- HEADER/SIGNATURE REGEX --------
    let order_invariant_header_regex_raw = `(((\\n|^)(((from):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>)?|(subject:[a-zA-Z 0-9]+)?|((to):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>)?)(\\r))+)\\n`;
    let sig_regex = `\r\ndkim-signature:(${key_chars}=${catch_all_without_semicolon}+; )+bh=${base_64}+; `;
    // let full_header_regex = order_invariant_header_regex_raw + sig_regex;
    // let raw_regex = order_invariant_header_regex_raw;
    // let regex = regexToMinDFASpec(raw_regex) + sig_regex;
    // console.log(format_regex_printable(sig_regex));
    // -------- SUBJECT REGEXES --------
    // This raw subject line (with \\ replaced with \) can be put into regexr.com to test new match strings and sanity check that it works
    // TODO: Other valid chars in email addresses: #$%!^/&*, outlined at https://ladedu.com/valid-characters-for-email-addresses-the-complete-list/ and in the RFC
    // -- SEND SPECIFIC REGEXES --
    // let send_specific_raw_subject_regex = `((\r\n)|^)subject:[Ss]end (\$)?[0-9]+(.[0-9]+)? [a-zA-Z]+ to (${email_address_regex}|0x[0-9a-fA_F]+)\r\n`;
    // let raw_subject_regex = `((\r\n)|^)subject:[a-zA-Z]+ (\\$)?[0-9]+(.[0-9]+)? [a-zA-Z]+ to (([a-zA-Z0-9._%\\+-=]+@[a-zA-Z0-9.-]+)|0x[0-9]+)\r\n`;
    // Input: ((\\r\\n)|^)subject:[Ss]end (\$)?[0-9]+(.[0-9]+)? [a-zA-Z]+ to (([a-zA-Z0-9._%\+-=]+@[a-zA-Z0-9.-]+)|0x[0-9]+)\\r\\n
    // This can be pasted into the first line of https://zkregex.com/min_dfa (after replacing \\ -> \)
    // ((\\r\\n)|\^)subject:[Ss]end (\$)?[0-9]+(\.[0-9])? (ETH|DAI|USDC|eth|usdc|dai) to (([a-zA-Z0-9\._%\+-]+@[a-zA-Z0-9\.-]+.[a-zA-Z0-9]+)|0x[0-9]+)\\r\\n
    // console.log(raw_subject_regex);
    // -- GENERIC SUBJECT COMMANDS --
    let raw_subject_regex = `((\r\n)|^)subject:[a-zA-Z]+ (\\$)?[0-9]+(.[0-9]+)? [a-zA-Z]+ to (${email_address_regex}|0x[0-9a-fA_F]+)\r\n`;
    // -------- OTHER FIELD REGEXES --------
    let raw_from_regex = `(\r\n|^)from:([A-Za-z0-9 _.,"@-]+)<[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.-]+>\r\n`;
    // let message_id_regex = `(\r\n|^)message-id:<[=@.\\+_-a-zA-Z0-9]+>\r\n`;
    // -------- TWITTER BODY REGEX ---------
    // let regex = STRING_PRESELECTOR + `${word_char}+`;
    // ---------- DEPRECATAED REGEXES ----------
    // let order_invariant_header_regex_raw = `(((\\n|^)(((from):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>)?|(subject:[a-zA-Z 0-9]+)?|((to):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>)?)(\\r))+)`;
    // let order_invariant_full_regex_raw = `(dkim-signature:((a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z)=(0|1|2|3|4|5|6|7|8|9|a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|!|"|#|$|%|&|\'|\\(|\\)|\\*|\\+|,|-|.|/|:|<|=|>|\\?|@|[|\\\\|]|^|_|\`|{|\\||}|~| |\t|\n|\r|\x0B|\f)+; ))?)(\\r))+` // Uses a-z syntax instead of | for each char
    // let old_regex = '(\r\n|\x80)(to|from):([A-Za-z0-9 _."@-]+<)?[a-zA-Z0-9_.-]+@[a-zA-Z0-9_.]+>?\r\n';
    // let regex = `(\n|^)(to|from):((${email_chars}|"|@| )+<)?(${email_chars})+@(${email_chars})+>?\r`;
    // let regex = `(\r\n|^)(to|from):((${email_chars}|"|@| )+<)?(${email_chars})+@(${email_chars})+>?\r\n`;
    // 'dkim-signature:((a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z)=(0|1|2|3|4|5|6|7|8|9|a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|!|"|#|$|%|&|\'|\\(|\\)|\\*|\\+|,|-|.|/|:|<|=|>|\\?|@|[|\\\\|]|^|_|`|{|\\||}|~| |\t|\n|\r|\x0B|\f)+; )+bh=(a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|0|1|2|3|4|5|6|7|8|9|\\+|/|=)+; '
    // let regex = 'hello(0|1|2|3|4|5|6|7|8|9)+world';
    // --------- FINAL CONVERSION ---------
    //   console.log(format_regex_printable(raw_subject_regex));
    // console.log('regex: ', base_64, '\n')
    let regex = regexToMinDFASpec("1=(a|b) (2=(b|c)+ )+d");
    // console.log(format_regex_printable(regex));
    console.log('expanded regex: ', regex);
    return regex;
}
// Escapes and prints regexes (might be buggy)
function format_regex_printable(s) {
    const escaped_string_json = JSON.stringify(s);
    const escaped_string = escaped_string_json.slice(1, escaped_string_json.length - 1);
    return escaped_string
        .replaceAll("\\\\\\\\", "\\")
        .replaceAll("\\\\", "\\")
        .replaceAll("\\|", "\\\\|")
        .replaceAll("/", "\\/")
        .replaceAll("\u000b", "\\♥")
        .replaceAll("|[|", "|\\[|")
        .replaceAll("|]|", "|\\]|")
        .replaceAll("|.|", "|\\.|")
        .replaceAll("|$|", "|\\$|")
        .replaceAll("|^|", "|\\^|");
    //   let escaped = escape_whitespace(escape_whitespace(s.replaceAll("\\\\", "ZZZZZZZ")));
    //   let fixed = escaped.replaceAll("\\(", "(").replaceAll("\\)", ")").replaceAll("\\+", "+").replaceAll("\\*", "*").replaceAll("\\?", "?");
}
// Note that this is not complete and very case specific i.e. can only handle a-z and a-f, and not a-c.
// This function expands [] sections to convert values for https://zkregex.com/min_dfa
// The input is a regex with [] and special characters (i.e. the first line of min_dfa tool)
// The output is expanded regexes without any special characters
function regexToMinDFASpec(str) {
    // Replace all A-Z with A2Z etc
    // TODO: Upstream this to min_dfa
    let combined_nosep = str
        .replaceAll("A-Z", A2Z_nosep)
        .replaceAll("a-z", a2z_nosep)
        .replaceAll("A-F", A2F_nosep)
        .replaceAll("a-f", a2f_nosep)
        .replaceAll("0-9", r0to9_nosep)
        .replaceAll("\\w", A2Z_nosep + r0to9_nosep + a2z_nosep + "_")
        .replaceAll("\\d", r0to9_nosep)
        .replaceAll("\\s", slash_s);
    // .replaceAll("\\w", A2Z_nosep + r0to9_nosep + a2z_nosep); // I think that there's also an underscore here
    function addPipeInsideBrackets(str) {
        let result = "";
        let insideBrackets = false;
        for (let i = 0; i < str.length; i++) {
            if (str[i] === "[") {
                result += str[i];
                insideBrackets = true;
                continue;
            }
            else if (str[i] === "]") {
                insideBrackets = false;
            }
            let str_to_add = str[i];
            if (str[i] === "\\") {
                i++;
                str_to_add += str[i];
            }
            result += insideBrackets ? "|" + str_to_add : str_to_add;
        }
        return result.replaceAll("[|", "[").replaceAll("[", "(").replaceAll("]", ")");
    }
    //   function makeCurlyBracesFallback(str) {
    //     let result = "";
    //     let insideBrackets = false;
    //     for (let i = 0; i < str.length; i++) {
    //       if (str[i] === "{") {
    //         result += str[i];
    //         insideBrackets = true;
    //         continue;
    //       } else if (str[i] === "}") {
    //         insideBrackets = false;
    //       }
    //       result += insideBrackets ? "|" + str[i] : str[i];
    //     }
    //     return result.replaceAll("[|", "[").replaceAll("[", "(").replaceAll("]", ")");
    //   }
    function checkIfBracketsHavePipes(str) {
        let result = true;
        let insideBrackets = false;
        let insideParens = 0;
        let indexAt = 0;
        for (let i = 0; i < str.length; i++) {
            if (indexAt >= str.length)
                break;
            if (str[indexAt] === "[") {
                insideBrackets = true;
                indexAt++;
                continue;
            }
            else if (str[indexAt] === "]") {
                insideBrackets = false;
            }
            if (str[indexAt] === "(") {
                insideParens++;
            }
            else if (str[indexAt] === ")") {
                insideParens--;
            }
            if (insideBrackets) {
                if (str[indexAt] === "|") {
                    indexAt++;
                }
                else {
                    result = false;
                    return result;
                }
            }
            if (!insideParens && str[indexAt] === "|") {
                console.log("Error: | outside of parens!");
            }
            if (str[indexAt] === "\\") {
                indexAt++;
            }
            indexAt++;
        }
        return result;
    }
    let combined;
    if (!checkIfBracketsHavePipes(combined_nosep)) {
        // console.log("Adding pipes within brackets between everything!");
        combined = addPipeInsideBrackets(combined_nosep);
        if (!checkIfBracketsHavePipes(combined)) {
            console.log("Did not add brackets correctly!");
        }
    }
    else {
        combined = combined_nosep;
    }
    return combined;
}
function toNature(col) {
    var i, j, base = "ABCDEFGHIJKLMNOPQRSTUVWXYZ", result = 0;
    if ("1" <= col[0] && col[0] <= "9") {
        result = parseInt(col, 10);
    }
    else {
        for (i = 0, j = col.length - 1; i < col.length; i += 1, j -= 1) {
            result += Math.pow(base.length, j) * (base.indexOf(col[i]) + 1);
        }
    }
    return result;
}
function printGraphForRegex(regex) {
    let nfa = regexToNfa(regex);
    let dfa = minDfa(nfaToDfa(nfa));
    var i, j, states = {}, nodes = [], stack = [dfa], symbols = [], top;
    while (stack.length > 0) {
        top = stack.pop();
        if (!states.hasOwnProperty(top.id)) {
            states[top.id] = top;
            top.nature = toNature(top.id);
            nodes.push(top);
            for (i = 0; i < top.edges.length; i += 1) {
                if (top.edges[i][0] !== "ϵ" && symbols.indexOf(top.edges[i][0]) < 0) {
                    symbols.push(top.edges[i][0]);
                }
                stack.push(top.edges[i][1]);
            }
        }
    }
    nodes.sort(function (a, b) {
        return a.nature - b.nature;
    });
    symbols.sort();
    let graph = [];
    for (let i = 0; i < nodes.length; i += 1) {
        let curr = {};
        curr.type = nodes[i].type;
        curr.edges = {};
        for (let j = 0; j < symbols.length; j += 1) {
            if (nodes[i].trans.hasOwnProperty(symbols[j])) {
                curr.edges[symbols[j]] = nodes[i].trans[symbols[j]].nature - 1;
            }
        }
        graph[nodes[i].nature - 1] = curr;
    }
    console.log(JSON.stringify(graph));
    return JSON.stringify(graph);
}
let regex = test_regex();
// console.log('regex:\n', regex);
const graphJson = printGraphForRegex(regex);
const circuit = genCircomAllstr(JSON.parse(graphJson), 'circuit');
console.log(circuit);
if (typeof require === "function") {
    exports.regexToMinDFASpec = regexToMinDFASpec;
    exports.toNature = toNature;
}
// ------------------------------------------
// function genCircomAllstr(graph_json, template_name) {
//   Set.prototype.isSuperset = function (subset) {
//     if (this.size === 0) {
//         return false;
//     }
//     for (var elem of subset) {
//         if (!this.has(elem)) {
//             return false;
//         }
//     }
//     return true;
//   }
//   Set.prototype.difference = function (setB) {
//     for (let elem of setB) {
//         this.delete(elem)
//     }
//   }
//   const N = graph_json.length;
//   // const graph = Array(N).fill({});
//   const rev_graph = [];
//   const to_init_graph = [];
//   let init_going_state = null;
//   for (let i = 0; i < N; i++) {
//       rev_graph.push({});
//       to_init_graph.push([]);
//   }
//   let accept_nodes = new Set();
//   for (let i = 0; i < N; i++) {
//       for (let k in graph_json[i]["edges"]) {
//           const v = graph_json[i]["edges"][k];
//           rev_graph[v][i] = Array.from(JSON.parse(k)).map(c => c.charCodeAt());
//           if (i === 0) {
//               const index = rev_graph[v][i].indexOf(94);
//               if (index !== -1) {
//                   init_going_state = v;
//                   rev_graph[v][i][index] = 128;
//               }
//               for (let j = 0; j < rev_graph[v][i].length; j++) {
//                   if (rev_graph[v][i][j] == 128) {
//                       continue;
//                   }
//                   to_init_graph[v].push(rev_graph[v][i][j]);
//               }
//           }
//       }
//       if (graph_json[i]["type"] == "accept") {
//           accept_nodes.add(i);
//       }
//   }
//   if (init_going_state !== null) {
//       for (const [going_state, chars] of Object.entries(to_init_graph)) {
//           if (chars.length === 0) {
//               continue;
//           }
//           if (rev_graph[going_state][init_going_state] == null) {
//               rev_graph[going_state][init_going_state] = [];
//           }
//           rev_graph[going_state][init_going_state] = rev_graph[going_state][init_going_state].concat(chars);
//       }
//   }
//   if (accept_nodes[0] === null) {
//       throw new Error("accept node must not be 0");
//   }
//   accept_nodes = [...accept_nodes];
//   if (accept_nodes.length !== 1) {
//       throw new Error("the size of accept nodes must be one");
//   }
//   let eq_i = 0;
//   let lt_i = 0;
//   let and_i = 0;
//   let multi_or_i = 0;
//   let lines = [];
//   lines.push("\tfor (var i = 0; i < num_bytes; i++) {");
//   const uppercase = new Set(Array.from("ABCDEFGHIJKLMNOPQRSTUVWXYZ").map(c => c.charCodeAt()));
//   const lowercase = new Set(Array.from("abcdefghijklmnopqrstuvwxyz").map(c => c.charCodeAt()));
//   const digits = new Set(Array.from("0123456789").map(c => c.charCodeAt()));
//   const symbols1 = new Set([":", ";", "<", "=", ">", "?", "@"].map(c => c.charCodeAt()));
//   const symbols2 = new Set(["[", "\\", "]", "^", "_", "`"].map(c => c.charCodeAt()));
//   const symbols3 = new Set(["{", "|", "}", "~"].map(c => c.charCodeAt()));
//   lines.push(`\t\tstate_changed[i] = MultiOR(${N - 1});`);
//   for (let i = 1; i < N; i++) {
//       const outputs = [];
//       let is_negates = [];
//       for (let prev_i of Object.keys(rev_graph[i])) {
//           const k = rev_graph[i][prev_i];
//           const eq_outputs = [];
//           let vals = new Set(k);
//           // let is_negate = false;
//           if (vals.has(0xff)) {
//               vals.delete(0xff);
//               is_negates.push(true);
//           } else {
//               is_negates.push(false);
//           }
//           if (vals.size === 0) {
//               continue;
//           }
//           if (is_negates[is_negates.length - 1] === true) {
//               for (let another_i = 1; another_i < N; another_i++) {
//                   if (i === another_i) {
//                       continue;
//                   }
//                   if (rev_graph[another_i][prev_i] === null) {
//                       continue;
//                   }
//                   const another_vals = new Set(rev_graph[another_i][prev_i]);
//                   if (another_vals.size === 0) {
//                       continue;
//                   }
//                   for (let another_val of another_vals) {
//                       vals.add(another_val);
//                   }
//               }
//           }
//           const min_maxs = [];
//           for (let subsets of [
//               [digits, 47, 58],
//               [symbols1, 57, 65],
//               [uppercase, 64, 91],
//               [symbols2, 90, 97],
//               [lowercase, 96, 123],
//               [symbols3, 122, 127]
//           ]) {
//               const subset = subsets[0];
//               const min = subsets[1];
//               const max = subsets[2];
//               if (vals.isSuperset(subset)) {
//                   vals.difference(subset);
//                   if (min_maxs.length == 0) {
//                       min_maxs.push([min, max]);
//                   } else {
//                       const last = min_maxs[min_maxs.length - 1];
//                       if (last[1] - 1 == min) {
//                           min_maxs[min_maxs.length - 1][1] = max;
//                       } else {
//                           min_maxs.push([min, max]);
//                       }
//                   }
//               }
//           }
//           for (let min_max of min_maxs) {
//               lines.push(`\t\tlt[${lt_i}][i] = LessThan(8);`);
//               lines.push(`\t\tlt[${lt_i}][i].in[0] <== ${min_max[0]};`);
//               lines.push(`\t\tlt[${lt_i}][i].in[1] <== in[i];`);
//               lines.push(`\t\tlt[${lt_i + 1}][i] = LessThan(8);`);
//               lines.push(`\t\tlt[${lt_i + 1}][i].in[0] <== in[i];`);
//               lines.push(`\t\tlt[${lt_i + 1}][i].in[1] <== ${min_max[1]};`);
//               lines.push(`\t\tand[${and_i}][i] = AND();`);
//               lines.push(`\t\tand[${and_i}][i].a <== lt[${lt_i}][i].out;`);
//               lines.push(`\t\tand[${and_i}][i].b <== lt[${lt_i + 1}][i].out;`);
//               eq_outputs.push(['and', and_i]);
//               lt_i += 2
//               and_i += 1
//           }
//           for (let code of vals) {
//               lines.push(`\t\teq[${eq_i}][i] = IsEqual();`);
//               lines.push(`\t\teq[${eq_i}][i].in[0] <== in[i];`);
//               lines.push(`\t\teq[${eq_i}][i].in[1] <== ${code};`);
//               eq_outputs.push(['eq', eq_i]);
//               eq_i += 1
//           }
//           lines.push(`\t\tand[${and_i}][i] = AND();`);
//           lines.push(`\t\tand[${and_i}][i].a <== states[i][${prev_i}];`);
//           if (eq_outputs.length === 1) {
//               lines.push(`\t\tand[${and_i}][i].b <== ${eq_outputs[0][0]}[${eq_outputs[0][1]}][i].out;`);
//           } else if (eq_outputs.length > 1) {
//               lines.push(`\t\tmulti_or[${multi_or_i}][i] = MultiOR(${eq_outputs.length});`);
//               for (let output_i = 0; output_i < eq_outputs.length; output_i++) {
//                   lines.push(`\t\tmulti_or[${multi_or_i}][i].in[${output_i}] <== ${eq_outputs[output_i][0]}[${eq_outputs[output_i][1]}][i].out;`);
//               }
//               lines.push(`\t\tand[${and_i}][i].b <== multi_or[${multi_or_i}][i].out;`);
//               multi_or_i += 1
//           }
//           outputs.push(and_i);
//           and_i += 1;
//       }
//       if (outputs.length === 1) {
//           if (is_negates[0]) {
//               lines.push(`\t\tstates[i+1][${i}] <== 1 - and[${outputs[0]}][i].out;`);
//           } else {
//               lines.push(`\t\tstates[i+1][${i}] <== and[${outputs[0]}][i].out;`);
//           }
//       } else if (outputs.length > 1) {
//           lines.push(`\t\tmulti_or[${multi_or_i}][i] = MultiOR(${outputs.length});`);
//           for (let output_i = 0; output_i < outputs.length; output_i++) {
//               if (is_negates[output_i]) {
//                   lines.push(`\t\tmulti_or[${multi_or_i}][i].in[${output_i}] <== 1 - and[${outputs[output_i]}][i].out;`);
//               } else {
//                   lines.push(`\t\tmulti_or[${multi_or_i}][i].in[${output_i}] <== and[${outputs[output_i]}][i].out;`);
//               }
//           }
//           lines.push(`\t\tstates[i+1][${i}] <== multi_or[${multi_or_i}][i].out;`);
//           multi_or_i += 1
//       }
//       lines.push(`\t\tstate_changed[i].in[${i - 1}] <== states[i+1][${i}];`);
//   }
//   lines.push(`\t\tstates[i+1][0] <== 1 - state_changed[i].out;`);
//   lines.push("\t}");
//   const declarations = [];
//   declarations.push(`pragma circom 2.1.5;\n`);
//   declarations.push(`include "zk-regex-circom/circuits/regex_helpers.circom";\n`);
//   // declarations.push(`pragma circom 2.1.5;\ninclude "@zk-email/circuits/regexes/regex_helpers.circom";\n`);
//   declarations.push(`template ${template_name}(msg_bytes) {`);
//   declarations.push(`\tsignal input msg[msg_bytes];`);
//   declarations.push(`\tsignal output out;\n`);
//   declarations.push(`\tvar num_bytes = msg_bytes+1;`);
//   declarations.push(`\tsignal in[num_bytes];`);
//   declarations.push(`\tin[0]<==128;`);
//   declarations.push(`\tfor (var i = 0; i < msg_bytes; i++) {`);
//   declarations.push(`\t\tin[i+1] <== msg[i];`);
//   declarations.push(`\t}\n`);
//   if (eq_i > 0) {
//       declarations.push(`\tcomponent eq[${eq_i}][num_bytes];`);
//   }
//   if (lt_i > 0) {
//       declarations.push(`\tcomponent lt[${lt_i}][num_bytes];`);
//   }
//   if (and_i > 0) {
//       declarations.push(`\tcomponent and[${and_i}][num_bytes];`);
//   }
//   if (multi_or_i > 0) {
//       declarations.push(`\tcomponent multi_or[${multi_or_i}][num_bytes];`);
//   }
//   declarations.push(`\tsignal states[num_bytes+1][${N}];`);
//   declarations.push(`\tcomponent state_changed[num_bytes];`);
//   declarations.push("");
//   const init_code = [];
//   init_code.push(`\tstates[0][0] <== 1;`);
//   init_code.push(`\tfor (var i = 1; i < ${N}; i++) {`);
//   init_code.push(`\t\tstates[0][i] <== 0;`);
//   init_code.push("\t}");
//   init_code.push("");
//   lines = declarations.concat(init_code).concat(lines);
//   const accept_node = accept_nodes[0];
//   const accept_lines = [""];
//   accept_lines.push("\tcomponent final_state_result = MultiOR(num_bytes+1);");
//   accept_lines.push("\tfor (var i = 0; i <= num_bytes; i++) {");
//   accept_lines.push(`\t\tfinal_state_result.in[i] <== states[i][${accept_node}];`);
//   accept_lines.push("\t}");
//   accept_lines.push("\tout <== final_state_result.out;");
//   lines = lines.concat(accept_lines);
//   let string = lines.reduce((res, line) => res + line + "\n", "");
//   return string;
// }
function genCircomAllstr(graph_json, template_name) {
    Set.prototype.isSuperset = function (subset) {
        if (this.size === 0) {
            return false;
        }
        for (var elem of subset) {
            if (!this.has(elem)) {
                return false;
            }
        }
        return true;
    };
    Set.prototype.difference = function (setB) {
        for (let elem of setB) {
            this.delete(elem);
        }
    };
    const N = graph_json.length;
    // const graph = Array(N).fill({});
    const rev_graph = [];
    const to_init_graph = [];
    let init_going_state = null;
    for (let i = 0; i < N; i++) {
        rev_graph.push({});
        to_init_graph.push([]);
    }
    let accept_nodes = new Set();
    for (let i = 0; i < N; i++) {
        for (let k in graph_json[i]["edges"]) {
            const v = graph_json[i]["edges"][k];
            rev_graph[v][i] = Array.from(JSON.parse(k)).map(c => c.charCodeAt());
            if (i === 0) {
                const index = rev_graph[v][i].indexOf(94);
                if (index !== -1) {
                    init_going_state = v;
                    rev_graph[v][i][index] = 128;
                }
                for (let j = 0; j < rev_graph[v][i].length; j++) {
                    if (rev_graph[v][i][j] == 128) {
                        continue;
                    }
                    to_init_graph[v].push(rev_graph[v][i][j]);
                }
            }
        }
        if (graph_json[i]["type"] == "accept") {
            accept_nodes.add(i);
        }
    }
    if (init_going_state !== null) {
        for (const [going_state, chars] of Object.entries(to_init_graph)) {
            if (chars.length === 0) {
                continue;
            }
            if (rev_graph[going_state][init_going_state] == null) {
                rev_graph[going_state][init_going_state] = [];
            }
            rev_graph[going_state][init_going_state] = rev_graph[going_state][init_going_state].concat(chars);
        }
    }
    if (accept_nodes[0] === null) {
        throw new Error("accept node must not be 0");
    }
    accept_nodes = [...accept_nodes];
    if (accept_nodes.length !== 1) {
        throw new Error("the size of accept nodes must be one");
    }
    let eq_i = 0;
    let lt_i = 0;
    let and_i = 0;
    let multi_or_i = 0;
    let lines = [];
    lines.push("\tfor (var i = 0; i < num_bytes; i++) {");
    const uppercase = new Set(Array.from("ABCDEFGHIJKLMNOPQRSTUVWXYZ").map(c => c.charCodeAt()));
    const lowercase = new Set(Array.from("abcdefghijklmnopqrstuvwxyz").map(c => c.charCodeAt()));
    const digits = new Set(Array.from("0123456789").map(c => c.charCodeAt()));
    const symbols1 = new Set([":", ";", "<", "=", ">", "?", "@"].map(c => c.charCodeAt()));
    const symbols2 = new Set(["[", "\\", "]", "^", "_", "`"].map(c => c.charCodeAt()));
    const symbols3 = new Set(["{", "|", "}", "~"].map(c => c.charCodeAt()));
    lines.push(`\t\tstate_changed[i] = MultiOR(${N - 1});`);
    for (let i = 1; i < N; i++) {
        const outputs = [];
        // let is_negates = [];
        for (let prev_i of Object.keys(rev_graph[i])) {
            const k = rev_graph[i][prev_i];
            const eq_outputs = [];
            let vals = new Set(k);
            let is_negate = false;
            if (vals.has(0xff)) {
                vals.delete(0xff);
                is_negate = true;
            }
            if (vals.size === 0) {
                continue;
            }
            if (is_negate === true) {
                for (let another_i = 1; another_i < N; another_i++) {
                    if (i === another_i) {
                        continue;
                    }
                    if (rev_graph[another_i][prev_i] === null) {
                        continue;
                    }
                    const another_vals = new Set(rev_graph[another_i][prev_i]);
                    if (another_vals.size === 0) {
                        continue;
                    }
                    for (let another_val of another_vals) {
                        vals.add(another_val);
                    }
                }
            }
            const min_maxs = [];
            for (let subsets of [
                [digits, 47, 58],
                [symbols1, 57, 65],
                [uppercase, 64, 91],
                [symbols2, 90, 97],
                [lowercase, 96, 123],
                [symbols3, 122, 127]
            ]) {
                const subset = subsets[0];
                const min = subsets[1];
                const max = subsets[2];
                if (vals.isSuperset(subset)) {
                    vals.difference(subset);
                    if (min_maxs.length == 0) {
                        min_maxs.push([min, max]);
                    }
                    else {
                        const last = min_maxs[min_maxs.length - 1];
                        if (last[1] - 1 == min) {
                            min_maxs[min_maxs.length - 1][1] = max;
                        }
                        else {
                            min_maxs.push([min, max]);
                        }
                    }
                }
            }
            for (let min_max of min_maxs) {
                lines.push(`\t\tlt[${lt_i}][i] = LessThan(8);`);
                lines.push(`\t\tlt[${lt_i}][i].in[0] <== ${min_max[0]};`);
                lines.push(`\t\tlt[${lt_i}][i].in[1] <== in[i];`);
                lines.push(`\t\tlt[${lt_i + 1}][i] = LessThan(8);`);
                lines.push(`\t\tlt[${lt_i + 1}][i].in[0] <== in[i];`);
                lines.push(`\t\tlt[${lt_i + 1}][i].in[1] <== ${min_max[1]};`);
                lines.push(`\t\tand[${and_i}][i] = AND();`);
                lines.push(`\t\tand[${and_i}][i].a <== lt[${lt_i}][i].out;`);
                lines.push(`\t\tand[${and_i}][i].b <== lt[${lt_i + 1}][i].out;`);
                eq_outputs.push(['and', and_i]);
                lt_i += 2;
                and_i += 1;
            }
            for (let code of vals) {
                lines.push(`\t\teq[${eq_i}][i] = IsEqual();`);
                lines.push(`\t\teq[${eq_i}][i].in[0] <== in[i];`);
                lines.push(`\t\teq[${eq_i}][i].in[1] <== ${code};`);
                eq_outputs.push(['eq', eq_i]);
                eq_i += 1;
            }
            lines.push(`\t\tand[${and_i}][i] = AND();`);
            lines.push(`\t\tand[${and_i}][i].a <== states[i][${prev_i}];`);
            if (eq_outputs.length === 1) {
                if (is_negate) {
                    lines.push(`\t\tand[${and_i}][i].b <== 1 - ${eq_outputs[0][0]}[${eq_outputs[0][1]}][i].out;`);
                }
                else {
                    lines.push(`\t\tand[${and_i}][i].b <== ${eq_outputs[0][0]}[${eq_outputs[0][1]}][i].out;`);
                }
            }
            else if (eq_outputs.length > 1) {
                lines.push(`\t\tmulti_or[${multi_or_i}][i] = MultiOR(${eq_outputs.length});`);
                for (let output_i = 0; output_i < eq_outputs.length; output_i++) {
                    lines.push(`\t\tmulti_or[${multi_or_i}][i].in[${output_i}] <== ${eq_outputs[output_i][0]}[${eq_outputs[output_i][1]}][i].out;`);
                }
                if (is_negate) {
                    lines.push(`\t\tand[${and_i}][i].b <== 1 - multi_or[${multi_or_i}][i].out;`);
                }
                else {
                    lines.push(`\t\tand[${and_i}][i].b <== multi_or[${multi_or_i}][i].out;`);
                }
                multi_or_i += 1;
            }
            outputs.push(and_i);
            and_i += 1;
        }
        if (outputs.length === 1) {
            lines.push(`\t\tstates[i+1][${i}] <== and[${outputs[0]}][i].out;`);
        }
        else if (outputs.length > 1) {
            lines.push(`\t\tmulti_or[${multi_or_i}][i] = MultiOR(${outputs.length});`);
            for (let output_i = 0; output_i < outputs.length; output_i++) {
                lines.push(`\t\tmulti_or[${multi_or_i}][i].in[${output_i}] <== and[${outputs[output_i]}][i].out;`);
            }
            lines.push(`\t\tstates[i+1][${i}] <== multi_or[${multi_or_i}][i].out;`);
            multi_or_i += 1;
        }
        lines.push(`\t\tstate_changed[i].in[${i - 1}] <== states[i+1][${i}];`);
    }
    lines.push(`\t\tstates[i+1][0] <== 1 - state_changed[i].out;`);
    lines.push("\t}");
    const declarations = [];
    declarations.push(`pragma circom 2.1.5;\n`);
    declarations.push(`include "zk-regex-circom/circuits/regex_helpers.circom";\n`);
    // declarations.push(`pragma circom 2.1.5;\ninclude "@zk-email/circuits/regexes/regex_helpers.circom";\n`);
    declarations.push(`template ${template_name}(msg_bytes) {`);
    declarations.push(`\tsignal input msg[msg_bytes];`);
    declarations.push(`\tsignal output out;\n`);
    declarations.push(`\tvar num_bytes = msg_bytes+1;`);
    declarations.push(`\tsignal in[num_bytes];`);
    declarations.push(`\tin[0]<==128;`);
    declarations.push(`\tfor (var i = 0; i < msg_bytes; i++) {`);
    declarations.push(`\t\tin[i+1] <== msg[i];`);
    declarations.push(`\t}\n`);
    if (eq_i > 0) {
        declarations.push(`\tcomponent eq[${eq_i}][num_bytes];`);
    }
    if (lt_i > 0) {
        declarations.push(`\tcomponent lt[${lt_i}][num_bytes];`);
    }
    if (and_i > 0) {
        declarations.push(`\tcomponent and[${and_i}][num_bytes];`);
    }
    if (multi_or_i > 0) {
        declarations.push(`\tcomponent multi_or[${multi_or_i}][num_bytes];`);
    }
    declarations.push(`\tsignal states[num_bytes+1][${N}];`);
    declarations.push(`\tcomponent state_changed[num_bytes];`);
    declarations.push("");
    const init_code = [];
    init_code.push(`\tstates[0][0] <== 1;`);
    init_code.push(`\tfor (var i = 1; i < ${N}; i++) {`);
    init_code.push(`\t\tstates[0][i] <== 0;`);
    init_code.push("\t}");
    init_code.push("");
    lines = declarations.concat(init_code).concat(lines);
    const accept_node = accept_nodes[0];
    const accept_lines = [""];
    accept_lines.push("\tcomponent final_state_result = MultiOR(num_bytes+1);");
    accept_lines.push("\tfor (var i = 0; i <= num_bytes; i++) {");
    accept_lines.push(`\t\tfinal_state_result.in[i] <== states[i][${accept_node}];`);
    accept_lines.push("\t}");
    accept_lines.push("\tout <== final_state_result.out;");
    lines = lines.concat(accept_lines);
    let string = lines.reduce((res, line) => res + line + "\n", "");
    return string;
}
Set.prototype.isSuperset = function (subset) {
    if (this.size === 0) {
        return false;
    }
    for (var elem of subset) {
        if (!this.has(elem)) {
            return false;
        }
    }
    return true;
};
Set.prototype.difference = function (setB) {
    for (let elem of setB) {
        this.delete(elem);
    }
};
//# sourceMappingURL=rawgex.js.map