export { parseRawRegex, generateMinDfaGraph, GraphTransition };
type GraphTransition = {
    type: string;
    transition: Record<string, number>;
};
/**
 * Parses the input raw regex into an expanded parsed form and displays the DFA.
 * Note: In order to specify some strings in regex, use \\ to escape \\s.
 * For instance, matching the literal + is represented as \\+, but \\r matches the literal \r character.
 * Matching \\ then an r as two characters would be represented as \\r in the JS string literal.
 *
 * @param {string} rawRegex The raw regex string to parse.
 * @returns {string} The expanded parsed regex.
 */
declare function parseRawRegex(rawRegex: string, logsEnabled?: boolean): string;
/**
 * Generates a minimum DFA graph from the given regular expression.
 *
 * @param {string} regex The (expanded) regular expression to generate the minimum DFA graph from.
 * @returns {string} The JSON representation of the minimum DFA graph.
 */
declare function generateMinDfaGraph(regex: string, logsEnabled?: boolean): string;
